/** Programa en C/C++ y OpenCV para :

 - Lee una imagen de entrada (está en tonos de gris)
 - La suaviza
 - La binariza con un umbral global
 - Se erosiona
 - Se dilata
 - Extrae los objetos presentes. Hará 10 imágenes de salida
   de los objetos más grandes presentes en la imagen e
   imprimirá en pantalla los momentos de Hu de cada objeto.

   13.07.2021
   Fraga
**/

#include "opencv2/core/utility.hpp"
#include "opencv2/core/mat.hpp"
#include "opencv2/imgproc.hpp"
#include "opencv2/imgcodecs.hpp"
#include <iostream>
#include <stdio.h>

using namespace cv;
using namespace std;

int do_comparison( const void *obja, const void *objb, void *ptr )
{
	int *p = (int *)ptr;

	return p[*(int *)obja] < p[*(int *)objb] ;	
}

int do_work( char *inputName  )
{
	char imgName[80]; 

	// Carga la imagen de entrada
    Mat Image1 = imread( inputName, IMREAD_GRAYSCALE );
	if( Image1.empty() ) {
		cerr << "Cannot read input image file: " << inputName << endl;
		return -1;
	}

	// Se suaviza la imagen 
	GaussianBlur( Image1, Image1, Size(5,5), 0, 0 );

	// Se umbraliza la imagen 
	threshold( Image1, Image1, 80, 255, THRESH_BINARY );
	Image1 = ~Image1;

	// Se erosiona y dilata
	Mat Element = getStructuringElement( MORPH_RECT, Size(3,3), Point(-1, -1) );

	Mat Image2;
	erode( Image1, Image2, Element, Point(-1,-1), 2 );
	dilate( Image2, Image1, Element, Point(-1,-1), 2 );

	imwrite( "bn.png", Image1 ); 

	// Se encuentran las componentes conectadas
    Mat Etiquetas, Estadisticas, Centroides;

	int n = connectedComponentsWithStats( Image1, Etiquetas, Estadisticas, Centroides, 8, CV_32S );

	int * tamanios = new int[n];
	int * etiquetas = new int[n-1];
	for( int i=1; i<n; i++ ) {
		etiquetas[i-1] = i;
		tamanios[i] = Estadisticas.at<int>( i, CC_STAT_AREA );
		// printf("%d %d\n", i, tamanios[i] );
	}
	qsort_r( etiquetas, n-1, sizeof(int), do_comparison, tamanios );
	/**
	for( int i=0; i<n-1; i++ ) {
		printf("%d %d\n", etiquetas[i], tamanios[ etiquetas[i] ] );
	}
	**/


	// Para los 10 objetos más grandes
	for( int i=1; i<n && i<=10; i++ ) {
		int m = etiquetas[i-1];
		sprintf( imgName, "img%03d.png", i );
		// Se calculan los momentos de Hu para el objeto
    	Image2 = Image2.zeros( Image2.size(), Image2.type() );

		int yini = Estadisticas.at<int>( m, CC_STAT_TOP );
		int yend = Estadisticas.at<int>( m, CC_STAT_HEIGHT );

		int xini = Estadisticas.at<int>( m, CC_STAT_LEFT );
		int xend = Estadisticas.at<int>( m, CC_STAT_WIDTH );

		for(int y=0;y<yend;y++) {
			int * ptr1 = Etiquetas.ptr<int>(yini+y) + xini; 
			unsigned char * ptr2 = Image2.ptr<uchar>(yini+y) + xini; 
			for(int x=0;x<xend;x++) {
				if( *ptr1 == m ) *ptr2 = 255;
				ptr1++;
				ptr2++;
			}
		}
		imwrite( imgName, Image2 );
		double hu[7];
		HuMoments( moments( Image2 ), hu );
		printf( "%d %lf %lf %lf %lf %lf %lf %lf\n", i, hu[0], hu[1], hu[2], hu[3], hu[4], hu[5], hu[6] ); 
	}

	delete[] tamanios;
	delete[] etiquetas;

	return EXIT_SUCCESS;
}

int main( int argc, char *argv[] )
{
	char inputName[128];

	if ( argc != 2 ) {
		cerr << "Args: input_image_name" << endl;
		return EXIT_FAILURE;
	}
	strncpy( inputName, argv[1], 127 );
	if ( do_work( inputName  ) != EXIT_SUCCESS ){
		cerr << "ERROR" << endl;
		exit( 1 );
	}
	return EXIT_SUCCESS;
}
